@@ -321,7 +321,7 @@ def ls(path, stat=False, assume_directory=False):
     if assume_directory:
         status = 1
     else:
-        status = is_file_or_dir(path)
+        status = is_file_or_dir(format(path, mgm))
     # Depending on status, return formatted path to file, directory contents, or raise
     if status == 0:
         raise RuntimeError('Path \'{0}\' does not exist'.format(path))
@@ -484,14 +484,10 @@ def ls_wildcard(pattern, stat=False):
             directories[:] = []
     return matches
 
-def hadd(src, dst, dry=False):
+def _hadd(root_files, dst, dry=False):
     """
-    Calls `ls_root` on `src` in order to be able to pass directories, then hadds.
-    Needs ROOT env to be callable.
+    Compiles and runs the hadd command
     """
-    root_files = ls_root(src)
-    if not len(root_files):
-        raise RuntimeError('src {0} yielded 0 root files'.format(src))
     cmd = ['hadd', '-f', dst] + root_files
     if dry:
         logger.warning('hadd command: ' + ' '.join(cmd))
@@ -508,6 +504,70 @@ def hadd(src, dst, dry=False):
     finally:
         debug(False)
 
+def hadd(src, dst, dry=False):
+    """
+    Calls `ls_root` on `src` in order to be able to pass directories, then hadds.
+    Needs ROOT env to be callable.
+    """
+    root_files = ls_root(src)
+    if not len(root_files):
+        raise RuntimeError('src {0} yielded 0 root files'.format(src))
+    _hadd(root_files, dst, dry=dry)
+
+def hadd_chunks(src, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dry=False):
+    """
+    Calls `ls_root` on `src` in order to be able to pass directories, then hadds.
+    Needs ROOT env to be callable.
+    """
+    root_files = ls_root(src)
+    if not len(root_files):
+        raise RuntimeError('src {0} yielded 0 root files'.format(src))
+    _hadd_chunks(root_files, dst, n_threads, chunk_size, tmpdir, dry)
+
+def _hadd_chunks(root_files, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dry=False):
+    """
+    Like hadd, but hadds a chunk of root files in threads to temporary files first,
+    then hadds the temporary files into the final root file.
+    The algorithm is recursive; if there are too many temporary files still, another intermediate
+    chunked hadd is performed.
+    """
+    if not len(root_files):
+        raise RuntimeError('src {0} yielded 0 root files'.format(src))
+    elif len(root_files) < chunk_size:
+        # No need for chunking. This should also be the final step of the recursion
+        _hadd(root_files, dst, dry=dry)
+        return
+
+    import math, uuid, shutil, multiprocessing as mp
+    n_chunks = int(math.ceil(len(root_files) / float(chunk_size)))
+
+    # Make a unique directory for temporary files
+    tmpdir = osp.join(tmpdir, 'tmphadd', str(uuid.uuid4()))
+    os.makedirs(tmpdir)
+
+    try:
+        debug(True)
+        chunk_rootfiles = []
+        p = mp.Pool(n_threads)
+        for i_chunk in range(n_chunks):
+            chunk = root_files[ i_chunk*chunk_size : (i_chunk+1)*chunk_size ]
+            chunk_dst = osp.join(tmpdir, 'chunk{0}.root'.format(i_chunk))
+            if dry:
+                logger.debug('hadding %s --> %s', ' '.join(chunk), chunk_dst)
+            else:
+                p.map(_hadd, (chunk, chunk_dst, dry))
+            chunk_rootfiles.append(chunk_dst)
+        p.close()
+        p.join()
+        # Merge the chunks into the final destination, potentially with another chunked merge
+        _hadd_chunks(chunk_rootfiles, dst, n_threads, chunk_size, tmpdir, dry)
+
+    finally:
+        logger.warning('Removing %s', tmpdir)
+        shutil.rmtree(tmpdir)
+        debug(False)
+
+
 # _______________________________________________________
 # Command line helpers
 
